////////////////////////////////////////////////////////////////////////
/// \file  ParticleListAction_service.h
/// \brief Use Geant4's user "hooks" to maintain a list of particles generated by Geant4.
///
/// \author  seligman@nevis.columbia.edu
////////////////////////////////////////////////////////////////////////
//
// accumulate a list of particles modeled by Geant4.
//

// Include guard
#ifndef PARTICLELISTACTION_SERVICE_H
#define PARTICLELISTACTION_SERVICE_H

#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceDeclarationMacros.h"

#include "lardataobj/Simulation/GeneratedParticleInfo.h"
#include "lardataobj/Simulation/ParticleAncestryMap.h"

#include "artg4tk/actionBase/EventActionBase.hh"
#include "artg4tk/actionBase/SteppingActionBase.hh"
#include "artg4tk/actionBase/TrackingActionBase.hh"

#include "nug4/ParticleNavigation/ParticleList.h"

#include "nusimdata/SimulationBase/MCParticle.h"
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nusimdata/SimulationBase/simb.h" // simb::GeneratedParticleIndex_t

#include "larcore/Geometry/Geometry.h"
#include "larcorealg/CoreUtils/ParticleFilters.h"

#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceDeclarationMacros.h"

#include "canvas/Persistency/Common/Assns.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Provenance/ProductID.h"

namespace art {
  class EDProductGetter;
}

#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Geant4/globals.hh"

class G4Event;
class G4Track;
class G4Step;

class TLorentzVector;

#include <map>
#include <memory>
#include <set>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

namespace larg4 {

  class ParticleListActionService : public artg4tk::EventActionBase,
                                    public artg4tk::TrackingActionBase,
                                    public artg4tk::SteppingActionBase {
  public:
    // Standard constructors and destructors;
    explicit ParticleListActionService(fhicl::ParameterSet const&);

    // UserActions method that we'll override, to obtain access to
    // Geant4's particle tracks and trajectories.
    void preUserTrackingAction(const G4Track*) override;
    void postUserTrackingAction(const G4Track*) override;
    void userSteppingAction(const G4Step*) override;

    /// Returns the index of primary truth (`sim::NoGeneratorIndex` if none).
    simb::GeneratedParticleIndex_t GetPrimaryTruthIndex(int trackId) const;

    // Called at the beginning of each event (note that this is after the
    // primaries have been generated and sent to the event manager)
    void beginOfEventAction(const G4Event*) override;

    // Called at the end of each event, right before GEANT's state switches
    // out of event processing and into closed geometry (last chance to access
    // the current event).
    void endOfEventAction(const G4Event*) override;

    // Set/get the current Art event
    void setInputCollections(std::vector<art::Handle<std::vector<simb::MCTruth>>> const& mclists)
    {
      fMCLists = &mclists;
    }

    void setPtrInfo(art::ProductID pid, art::EDProductGetter const* productGetter)
    {
      pid_ = pid;
      productGetter_ = productGetter;
    }

    std::unique_ptr<std::vector<simb::MCParticle>> ParticleCollection()
    {
      return std::move(partCol_);
    }
    std::unique_ptr<std::vector<simb::MCParticle>> DroppedParticleCollection()
    {
      return std::move(droppedPartCol_);
    }
    std::unique_ptr<sim::ParticleAncestryMap> DroppedTracksCollection()
    {
      return std::move(droppedCol_);
    }

    std::unique_ptr<art::Assns<simb::MCTruth, simb::MCParticle, sim::GeneratedParticleInfo>>
    AssnsMCTruthToMCParticle()
    {
      return std::move(tpassn_);
    }

    std::map<int, int> GetTargetIDMap() { return fTargetIDMap; }

    /// Grabs a particle filter
    void CreateParticleFilter(std::vector<std::string> keepParticlesInVolumes,
                              std::unique_ptr<util::PositionInVolumeFilter>& filter)
    {
      // if we don't have favourite volumes, don't even bother creating a filter
      std::set<std::string> vol_names(keepParticlesInVolumes.begin(), keepParticlesInVolumes.end());

      if (empty(vol_names)) filter = {};

      auto const& geom = *art::ServiceHandle<geo::Geometry const>();

      std::vector<std::vector<TGeoNode const*>> node_paths = geom.FindAllVolumePaths(vol_names);

      // collection of interesting volumes
      util::PositionInVolumeFilter::AllVolumeInfo_t GeoVolumePairs;
      GeoVolumePairs.reserve(node_paths.size()); // because we are obsessed

      //for each interesting volume, follow the node path and collect
      //total rotations and translations
      for (size_t iVolume = 0; iVolume < node_paths.size(); ++iVolume) {
        std::vector<TGeoNode const*> path = node_paths[iVolume];

        auto pTransl = new TGeoTranslation(0., 0., 0.);
        auto pRot = new TGeoRotation();
        for (TGeoNode const* node : path) {
          TGeoTranslation thistranslate(*node->GetMatrix());
          TGeoRotation thisrotate(*node->GetMatrix());
          pTransl->Add(&thistranslate);
          *pRot = *pRot * thisrotate;
        }

        // for some reason, pRot and pTransl don't have tr and rot bits set
        // correctly make new translations and rotations so bits are set correctly
        auto pTransl2 = new TGeoTranslation(
          pTransl->GetTranslation()[0], pTransl->GetTranslation()[1], pTransl->GetTranslation()[2]);
        double phi = 0., theta = 0., psi = 0.;
        pRot->GetAngles(phi, theta, psi);
        auto pRot2 = new TGeoRotation();
        pRot2->SetAngles(phi, theta, psi);

        auto pTransf = new TGeoCombiTrans(*pTransl2, *pRot2);
        GeoVolumePairs.emplace_back(node_paths[iVolume].back()->GetVolume(), pTransf);
      }

      filter = std::make_unique<util::PositionInVolumeFilter>(std::move(GeoVolumePairs));
    }
    void ParticleFilter() { CreateParticleFilter(fKeepParticlesInVolumes, fFilter); }
    void DroppedParticleFilter()
    {
      CreateParticleFilter(fKeepDroppedParticlesInVolumes, fDroppedFilter);
    }
    /// Return whether dropped particles are stored
    bool storeDropped() const { return fStoreDroppedMCParticles; }

  private:
    struct ParticleInfo_t {
      simb::MCParticle* particle = nullptr; ///< simple structure representing particle
      bool keepFullTrajectory = false;      ///< if there was decision to keep
      bool isInVolume = false;              ///< drop if not involume
      bool isDropped = false;               ///< dropped by a physics process

      /// Index of the particle in the original generator truth record.
      simb::GeneratedParticleIndex_t truthIndex = simb::NoGeneratedParticleIndex;

      /// Resets the information (does not release memory it does not own)
      void clear()
      {
        particle = nullptr;
        keepFullTrajectory = false;
        isInVolume = false;
        isDropped = false;
        truthIndex = simb::NoGeneratedParticleIndex;
      }

      /// Returns whether there is a particle
      bool hasParticle() const { return particle; }

      /// Returns whether there is a particle
      bool isPrimary() const { return simb::isGeneratedParticleIndex(truthIndex); }

      /// Returns the index of the particle in the generator truth record.
      simb::GeneratedParticleIndex_t truthInfoIndex() const { return truthIndex; }

      /// Print
      void print()
      {
        std::cout << "ParticleInfo_t: " << std::endl;
        std::cout << "  particle: " << particle << std::endl;
        std::cout << "  keepFullTrajectory: " << keepFullTrajectory << std::endl;
        std::cout << "  isInVolume: " << isInVolume << std::endl;
        std::cout << "  isDropped: " << isDropped << std::endl;
        std::cout << "  truthIndex: " << truthIndex << std::endl;
      }

    }; // ParticleInfo_t

    // Returns whether the particle was dropped
    bool isDropped(simb::MCParticle const* p);

    // Yields the ParticleList accumulated during the current event.
    sim::ParticleList&& YieldList();

    // Yields the (dropped) ParticleList accumulated during the current event.
    sim::ParticleList&& YieldDroppedList();

    // this method will loop over the fParentIDMap to get the
    // parentage of the provided trackid
    int GetParentage(int trackid) const;

    G4double fenergyCut;             ///< The minimum energy for a particle to
                                     ///< be included in the list.
    ParticleInfo_t fCurrentParticle; ///< information about the particle currently being simulated
                                     ///< for a single particle.
    sim::ParticleList fParticleList; ///< The accumulated particle information for
                                     ///< all particles in the event.
    G4bool fstoreTrajectories;       ///< Whether to store particle trajectories with each particle.
    std::vector<std::string>
      fkeepGenTrajectories;          ///< List of generators for which fstoreTrajectories applies.
                                     ///  if not provided and storeTrajectories is true, then all
                                     ///  trajectories for all generators will be stored. If
                                     ///  storeTrajectories is set to false, this list is ignored
                                     ///  and all additional trajectory points are not stored.
    std::map<int, int> fParentIDMap; ///< key is current track ID, value is parent ID
    std::map<int, int>
      fTargetIDMap; ///< key is original track ID, value is ID to assign for downstream objs (e.g. SimEdeps)
    int fCurrentTrackID;         ///< track ID of the current particle, set to eve ID
                                 ///< for EM shower particles
    mutable int fTrackIDOffset;  ///< offset added to track ids when running over
                                 ///< multiple MCTruth objects.
    bool fKeepEMShowerDaughters; ///< whether to keep EM shower secondaries, tertiaries, etc
    std::vector<std::string> fNotStoredPhysics; ///< Physics processes that will not be stored
    bool fkeepOnlyPrimaryFullTraj; ///< Whether to store trajectories only for primaries and
                                   ///  their descendants with MCTruth process = "primary"
    bool fSparsifyTrajectories;    ///< help reduce the number of trajectory points.
    double fSparsifyMargin;        ///< set the sparsification margin
    bool fKeepTransportation;      ///< tell whether or not to keep the transportation process
    bool fKeepSecondToLast; ///< tell whether or not to force keeping the second to last point

    std::vector<std::string>
      fKeepParticlesInVolumes; ///<Only write particles that have trajectories through these volumes

    std::vector<std::string> fKeepDroppedParticlesInVolumes; ///<Only write particles that have
                                                             ///<trajectories through these volumes
    bool
      fStoreDroppedMCParticles; ///< Whether to keep a `sim::MCParticle` list of dropped particles

    std::unique_ptr<sim::ParticleList>
      fdroppedParticleList; ///< The accumulated particle information for
                            ///< all (dropped) particles in the event.
    std::vector<art::Handle<std::vector<simb::MCTruth>>> const*
      fMCLists; ///< MCTruthCollection input lists

    /// Map: particle track ID -> index of primary information in MC truth.
    std::map<int, simb::GeneratedParticleIndex_t> fPrimaryTruthMap;

    /// Map: particle track ID -> index of primary parent in std::vector<simb::MCTruth> object
    std::map<int, size_t> fMCTIndexMap;

    /// Map: particle trakc ID -> boolean decision to keep or not full trajectory points
    std::map<int, bool> fMCTPrimProcessKeepMap;

    /// Map: MCTruthIndex -> generator, input label of generator and keepGenerator decision
    std::map<size_t, std::pair<std::string, G4bool>> fMCTIndexToGeneratorMap;

    /// Map: not stored process and counter
    std::unordered_map<std::string, int> fNotStoredCounterUMap;

    /// map <ParentID, set: list of track ids for which no MCParticle was created>
    std::map<int, std::set<int>> fdroppedTracksMap;

    std::unique_ptr<std::vector<simb::MCParticle>> partCol_;
    /// This collection will hold the MCParticleLite objects created from dropped particles
    std::unique_ptr<std::vector<simb::MCParticle>> droppedPartCol_;
    std::unique_ptr<sim::ParticleAncestryMap> droppedCol_;
    std::unique_ptr<art::Assns<simb::MCTruth, simb::MCParticle, sim::GeneratedParticleInfo>>
      tpassn_;
    art::ProductID pid_{art::ProductID::invalid()};
    art::EDProductGetter const* productGetter_{nullptr};
    std::unique_ptr<util::PositionInVolumeFilter> fFilter; ///< filter for particles to be kept
    std::unique_ptr<util::PositionInVolumeFilter>
      fDroppedFilter; ///< filter for dropped particles to be kept (if any)
    /// Adds a trajectory point to the current particle, and runs the filter
    void AddPointToCurrentParticle(TLorentzVector const& pos,
                                   TLorentzVector const& mom,
                                   std::string const& process);
  };

} // namespace larg4

DECLARE_ART_SERVICE(larg4::ParticleListActionService, LEGACY)

#endif // PARTICLELISTACTION_SERVICE_H
